from __future__ import unicode_literals

import datetime
import json
import smhasher

from collections import namedtuple
from admin_page_lock.models.base_model import BasePageLockModel
from admin_page_lock.settings import (
    TIMEOUT,
    REDIS_PREFIX,
    REDIS_SETTINGS,
)
from redis import StrictRedis
from redis.exceptions import RedisError


RedisSettings = namedtuple(
    'RedisSettings',
    ['host', 'port', 'password', 'timeout']
)


class RedisPageLockModel(BasePageLockModel):
    @classmethod
    def _get_page_reference(cls, req):
        # Include `url parameters` to `page reference` in case
        # `URL_IGNORE_PARAMETERS` is set up to `False`.
        page_parameters = cls._get_page_url_parameters(req)
        page_url = cls._get_page_url(req)

        # Encode `url_parameters` and `url`.
        hex_url = hex(
            smhasher.murmur3_x64_128(
                '{}{}'.format(
                    page_url,
                    page_parameters
                )
            )
        )

        # Return unique page reference that will be used as `redis key`.
        return '{}:{}'.format(
            REDIS_PREFIX,
            hex_url
        )

    @classmethod
    def _get_redis_client(cls):
        redis_settings = cls._get_redis_settings()

        redis_client = StrictRedis(
            host=redis_settings.host,
            port=redis_settings.port,
            password=redis_settings.password,
            socket_timeout=redis_settings.timeout
        )

        return redis_client

    @classmethod
    def _get_redis_settings(cls):
        # TODO(vstefka) add try block
        redis_settings = RedisSettings(
            REDIS_SETTINGS['host'],
            REDIS_SETTINGS['port'],
            REDIS_SETTINGS['password'],
            REDIS_SETTINGS['timeout']
        )

        return redis_settings

    @classmethod
    def deactivate(cls, page_settings):
        # Deactivate page connection by deleting stored data for current page
        # in `Redis`.
        redis_client = page_settings['redis_client']
        page_reference = page_settings['page_reference']

        redis_client.delete(page_reference)

    @classmethod
    def get_page_settings(cls, req):
        page_settings = super(RedisPageLockModel, cls).get_page_settings(req)

        page_settings['page_reference'] = cls._get_page_reference(req)
        page_settings['redis_client'] = cls._get_redis_client()

        return page_settings

    @classmethod
    def get_data(cls, page_settings):
        # Get data from `Redis` for page defined by `page_reference`.
        redis_client = page_settings['redis_client']
        page_reference = page_settings['page_reference']

        try:
            data = redis_client.get(page_reference)
            data = json.loads(data)
        except (
            IndexError,
            RedisError,
            TypeError
        ):
            return None

        data_to_return = {}
        for parameter_name, parameter_value in data.items():
            # Return parameters `locked_at` and `locked_out` as
            # `datetime.datetime` instances.
            if parameter_name in ['locked_at', 'locked_out']:
                parameter_value = datetime.datetime.strptime(
                    parameter_value,
                    '%Y-%m-%d-%H-%M-%S'
                )
            data_to_return.update({
                parameter_name: parameter_value
            })

        return data_to_return

    @classmethod
    def set_data(cls, page_settings, data):
        page_reference = page_settings['page_reference']
        redis_client = page_settings['redis_client']

        data_to_store = {}
        for parameter_name, parameter_value in data.items():
            # Store `datetime.datetime` data in string format in Redis.
            if parameter_name in ['locked_at', 'locked_out']:
                parameter_value = parameter_value.strftime('%Y-%m-%d-%H-%M-%S')

            data_to_store.update({
                parameter_name: parameter_value
            })

        try:
            redis_client.set(
                page_reference,
                json.dumps(data_to_store),
                TIMEOUT  # This deactives old records.
            )
        except RedisError:
            raise



from __future__ import unicode_literals

from django.contrib import admin
from admin_page_lock.models.database_model import DatabasePageLockModel


class DatabasePageLockModelAdmin(admin.ModelAdmin):
    fields = (
        'active',
        'user_reference',
        'locked',
        'url',
        'url_parameters',
    )
    list_display = (
        'url',
        'url_parameters',
        'active',
        'user_reference',
        'locked',
        'formated_locked_at',
        'formated_locked_out',
    )
    list_filter = (
        'active',
        'user_reference',
    )
    list_per_page = 20
    ordering = ('-locked_at',)

    def _get_formated_datetime(self, lock_datetime):
        return lock_datetime.strftime('%Y-%m-%d-%H-%M-%S')

    def formated_locked_at(self, obj):
        return self._get_formated_datetime(obj.locked_at)
    formated_locked_at.admin_order_field = 'locked_at'
    formated_locked_at.allow_tags = True
    formated_locked_at.short_description = DatabasePageLockModel._meta.get_field(  # noqa: E501
        'locked_at').verbose_name

    def formated_locked_out(self, obj):
        return self._get_formated_datetime(obj.locked_out)
    formated_locked_out.admin_order_field = 'locked_at'
    formated_locked_out.allow_tags = True
    formated_locked_out.short_description = DatabasePageLockModel._meta.get_field(  # noqa: E501
        'locked_out').verbose_name


admin.site.register(DatabasePageLockModel, DatabasePageLockModelAdmin)



from __future__ import unicode_literals

from django.conf.urls import url
from admin_page_lock.views import (
    ClosePageConnection,
    GetPageInfo,
    OpenPageConnection,
)

urlpatterns = [
    # Close Page Connection.
    url(
        r'^close_page_connection/$',
        ClosePageConnection.as_view(),
        name='page_lock_close_page_connection'
    ),
    # Get Page Info.
    url(
        r'^get_page_info/$',
        GetPageInfo.as_view(),
        name='page_lock_get_page_connection'
    ),
    # Open Page Connection.
    url(
        r'^open_page_connection/$',
        OpenPageConnection.as_view(),
        name='page_lock_open_page_connection'
    ),
]



from __future__ import unicode_literals

import json

from django.http import (
    HttpResponse,
)
from django.views.generic.base import View
from admin_page_lock.settings import (
    HANDLER_FUNCTION_CLOSE_PAGE_CONNECTION,
    HANDLER_FUNCTION_GET_PAGE_INFO,
    HANDLER_FUNCTION_OPEN_PAGE_CONNECTION
)
from admin_page_lock.utils import get_page_lock_classes


class BasePageView(View):
    HANDLER_FUNCTION = None

    def post(self, req, *args, **kwargs):
        # Get and initialize handler.
        # TODO(vstefka) type of handler can depend on current page as well.
        try:
            # Get and initialize handler.
            handler_class, model_class = get_page_lock_classes()
            handler = handler_class(req, model_class)

            # Get and run handler function to get response data.
            handler_function = getattr(handler, self.HANDLER_FUNCTION)
            response_data = handler_function(req, *args, **kwargs)
        except Exception as e:
            # Propagate error to django in order to be able to log it.
            raise e

        response = HttpResponse(
            json.dumps(response_data),
            content_type='application/json'
        )

        return response


class ClosePageConnection(BasePageView):
    """
    Call it when user is leaving the page.

    REQUEST:
     + csrf_token       CSRF token if `DISABLE_CRSF_TOKEN == True`;
     + url              url of locked page (with parameters);
     + user_reference   user reference visiting locked page.
    """
    HANDLER_FUNCTION = HANDLER_FUNCTION_CLOSE_PAGE_CONNECTION


class GetPageInfo(BasePageView):
    """
    Call it when user comes to the page at the first time. API then returns
    `RESPONSE` parameters based on `REQUEST` parameters and current`settings`.

    REQUEST:
     + csrf_token           CSRF token if `DISABLE_CRSF_TOKEN == True`;
     + url                  url of locked page (with parameters);
     + user_reference       user reference visiting locked page.
    RESPONSE:
     + is_locked            whether the page is locked (True/False);
     + locked_by            `user_reference` that locked the page;
     + page_lock_settings   current page lock type of the page (behaviou);
     + reconnect_in         number of seconds when page might be available.
    """
    HANDLER_FUNCTION = HANDLER_FUNCTION_GET_PAGE_INFO


class OpenPageConnection(BasePageView):
    """
    Call it when user wants to open/reopen page connection.

    REQUEST:
     + csrf_token           CSRF token if `DISABLE_CRSF_TOKEN == True`;
     + url                  url of locked page (with parameters);
     + user_reference       user reference visiting locked page.
    RESPONSE:
     + page_lock_settings   current lock page configuratino;
     + is_locked            whether page is locked (True/False);
     + reconnected          page is locked by same user;
     + reconnect_in         number of seconds when page might be available;
    """
    HANDLER_FUNCTION = HANDLER_FUNCTION_OPEN_PAGE_CONNECTION


# TODO(vstefka) think about migratting logic related to reopening from
# `OpenPageConnection` to new API `ReopenPageConnection`.



from __future__ import unicode_literals

import imp
import io
import os

from setuptools import find_packages, setup


def get_file_contents(filename):
    """Return file contents."""
    file_path = os.path.join(
        os.path.realpath(os.path.dirname(__file__)), filename)

    with io.open(file_path, encoding='utf-8') as fd:
        return fd.read().strip().replace('\r', '')


def get_module():
    """Return module."""
    init_path = os.path.join(
        os.path.realpath(os.path.join(os.path.dirname(__file__))),
        'admin_page_lock', '__init__.py')

    return imp.load_source('page_lock', init_path)


# Allow setup.py to be run from any path.
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    author='Vojtech Stefka',
    author_email='oss+djangopage@showmax.com',
    classifiers=[
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Framework :: Django :: 1.8',
        'Framework :: Django :: 1.11',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    description='Page Lock application prevents users from editing '
                'a page while it is being edited by someone else. '
                'The application is tailored to django admin implementation.',
    include_package_data=True,
    install_requires=[],
    keywords=['django', 'admin', 'locking', 'concurrency'],
    license='Apache License, Version 2.0',
    long_description=get_file_contents('README.md'),
    long_description_content_type='text/markdown',
    name=get_module().NAME,
    packages=find_packages(),
    platforms=['any'],
    version=get_module().VERSION,
    url='https://github.com/ShowMax/django-admin-page-lock',
)



from __future__ import unicode_literals

import json

from django.contrib import messages
from django.utils.translation import ugettext_lazy as _
from admin_page_lock.settings import (
    API_INTERVAL,
    CAN_OPEN_MORE_TABS,
    HANDLER_FUNCTION_CLOSE_PAGE_CONNECTION,
    HANDLER_FUNCTION_GET_PAGE_INFO,
    HANDLER_FUNCTION_OPEN_PAGE_CONNECTION
)
from admin_page_lock.utils import get_page_lock_classes


class BaseLockingMixin(object):
    lock_change_view = False
    lock_changelist_view = False

    def _add_extra_content(self, req, data):
        # Adding extra content.
        extra_context = {
            'page_lock_template_data': json.dumps(data),
            'page_lock_api_interval': int(API_INTERVAL)  # must be integer
        }

        if (
            data['is_locked'] and
            data['page_lock_settings']['user_reference'] != data['locked_by']  # noqa: E501
        ):
            # Adding message when page is locked.
            self._add_message(req, data)

        return extra_context

    def _add_message(self, req, data):
        # Adding message when page is locked.
        # TODO(vstefka) move message to settings.
        # TODO(vstefka) add setting attribute to hide the message.
        messages.warning(
            req,
            _('This page is locked by "{}".'.format(data['locked_by']))
        )

    @classmethod
    def _get_api_data(cls, req, handler_function_name):
        # Input argument `handler_function_name` must be one of defined in
        # settings.
        meta_name = '{}_data'.format(handler_function_name)
        if meta_name not in req.META:
            # Get and initialize handler.
            handler_class, model_class = get_page_lock_classes()
            lock_handler = handler_class(req, model_class)

            # Get and run handler function to get response data.
            handler_function = getattr(
                lock_handler,
                handler_function_name
            )
            req.META[meta_name] = handler_function(req)

        return req.META[meta_name]

    @classmethod
    def _get_page_info_data(cls, req):
        return cls._get_api_data(req, HANDLER_FUNCTION_GET_PAGE_INFO)

    @classmethod
    def _is_locked(cls, req):
        # Returns `True` if the current page is not locked by same user
        # otherwise returns `False`.
        result = cls._get_page_info_data(req)
        if (
            result['is_locked'] and
            result['page_lock_settings']['user_reference'] != result['locked_by']  # noqa: E501
        ):
            return True

        return False

    def _close_page_connection_data(self, req):
        return self._get_api_data(req, HANDLER_FUNCTION_CLOSE_PAGE_CONNECTION)

    def _open_page_connection_data(self, req):
        return self._get_api_data(req, HANDLER_FUNCTION_OPEN_PAGE_CONNECTION)

    def _redirect_to_homepage(self, data):
        # TODO(vstefka) this is not working!
        # If page is already locked by user then redirect him/her
        # to `homepage` defined in `settings`.
        if not CAN_OPEN_MORE_TABS and data['reconnected']:
            return True

        return False
